// Copyright 2021, CGVR, Contact: muehlenb@uni-bremen.de

#version 330 core
#define LIGHT_COUNT 100

#define M_E 2.71828182846
#define M_PI 3.14159265359
#define M_GAMMA 2.2

out vec4 FragColor;

// Position in view coordinates:
in vec4 mPos; 

// Normal in view coordinates:
in vec3 mNormal;
in mat3 TBN;

// Texture coordinates:
in vec2 texCoord; 

in vec4 mGlobalLightPoint;

in vec3 mLightDir;

in vec4 mLightPositions[4];
in vec4 mLightFragPosLightSpace[4];
uniform sampler2D shadowMapTexture[4];

// Material settings:
uniform vec4 color;
uniform vec4 emission;

uniform float metallic;
uniform float roughness;

float ambientLight = 0.0;

uniform sampler2D albedoTexture;
uniform sampler2D normalTexture;
uniform sampler2D metallicTexture;
uniform sampler2D roughnessTexture;

uniform bool useAlbedoTexture;
uniform bool useNormalTexture;
uniform bool useMetallicTexture;
uniform bool useRoughnessTexture;

/**
 * PBR LIGHTING FUNCTION FROM OLDER PROJECT 
 *
 * Schlicks approximation for fresnel factor:
 */

vec3 fresnel(float cosTheta, vec3 r0)
{
    return r0 + (1.0 - r0) * pow(1.0 - cosTheta, 5.0);
}  

float fresnel(float cosTheta, float r0)
{
    return r0 + (1.0 - r0) * pow(1.0 - cosTheta, 5.0);
}  

/**
 * PBR LIGHTING FUNCTION FROM OLDER PROJECT
 *
 * Trowbridge-Reitz GGX (Normal Distribution Function):
 */

float D_GGX(vec3 normal, vec3 halfVec, float pRoughness)
{
    float a = pRoughness * pRoughness;
    float a2 = a * a;
    float NdH = max(dot(normal, halfVec), 0.0);
    float NdH2 = NdH*NdH;
	float p = (NdH2 * (a2 - 1.0) + 1.0);
    return a2 / (M_PI * p * p);
}

/**
 * PBR LIGHTING FUNCTION FROM OLDER PROJECT
 *
 * Schlicks approximation of the smith equation:
 */

float G_Schlick(float NdV, float pRoughness)
{
    float r = (pRoughness + 1.0);
    float k = (r*r) / 0.7979;

    return NdV / (NdV * (1.0 - k) + k);
}

/**
 * PBR LIGHTING FUNCTION FROM OLDER PROJECT
 * 
 * Smith equation (G is divided into two components: light & view):
 */

float G_Smith(vec3 N, vec3 V, vec3 L, float pRoughness)
{
    float g1  = G_Schlick(max(dot(N, V), 0.0), pRoughness);
    float g2  = G_Schlick(max(dot(N, L), 0.0), pRoughness);
	
    return g1 * g2;
}

/**
 * PBR LIGHTING FUNCTION FROM OLDER PROJECT
 */

vec3 cookTorrance(vec3 normal, vec3 halfVec, vec3 toCamera, vec3 toLight, vec3 F, float pRoughness){
    // Berechne NDF und G Term:
    float NDF = D_GGX(normal, halfVec, pRoughness);       
    float G = G_Smith(normal, toCamera, toLight, pRoughness);   
        
    // Berechne die Cook-Torrance BRDF:
    vec3 nominator = NDF * G * F;
    float denominator = 4 * max(dot(normal, toCamera), 0.0)
            * max(dot(normal, toLight), 0.0) + 0.0005; 
                
    return nominator / denominator;  
}

float ShadowCalculation(int lightID)
{
    // perspective divide
    vec3 projCoords = mLightFragPosLightSpace[lightID].xyz / mLightFragPosLightSpace[lightID].w;
    projCoords = projCoords * 0.5 + 0.5;
			
	float shadow = 0.0;
	vec2 texelSize = 1.0 / textureSize(shadowMapTexture[lightID], 0);
	
	for(int x = -1; x <= 1; ++x)
	{
		for(int y = -1; y <= 1; ++y)
		{
			if(projCoords.x < 1 && projCoords.y < 1 && projCoords.x > 0 && projCoords.y > 0 && projCoords.z > 0 && projCoords.z < 1) {
				float closestDepth = texture(shadowMapTexture[lightID], projCoords.xy + vec2(x, y) * texelSize).r; 
				float currentDepth = projCoords.z - 0.0001;
	
				vec2 centerDist = vec2(0.5 - projCoords.x, 0.5 - projCoords.y);
				float fadeOutFactor = sqrt(max(0,1 - length(centerDist) * 2));
	
				shadow += (currentDepth > closestDepth  ? 1.0 : 0.0) * fadeOutFactor;
				
			}
		}    
	}
	shadow /= 9.0;
	
	return shadow;    
}

/**
 * PBR LIGHTING FUNCTION FROM OLDER PROJECT
 */

float LinearizeDepth(float depth)
{
    float z = depth * 2.0 - 1.0; // Back to NDC 
    return (2.0 * 0.02 * 50) / (50 + 0.02 - z * (50 - 0.2));
}

void getLighting(in vec4 pPos, in vec3 pNormal, in vec3 pAlbedo, in vec3 pEmission, in float pMetallic, in float pRoughness, out vec3 totalDiffuse, out vec3 totalSpecular){
    // Vectors to fragment and to camera in view space:
    vec3 toFragment = normalize(pPos.xyz);
    vec3 toCamera = -toFragment;
    
	vec3 rAlbedo = pAlbedo;
	float rMetallic = pMetallic;
	float rRoughness = pRoughness;
	vec3 rNormal = pNormal;
	
	if(useAlbedoTexture){
		/** Shadow Map debug:
		rAlbedo *= vec3(LinearizeDepth(texture(albedoTexture, texCoord).r) / 10);
		totalDiffuse = rAlbedo;
		totalSpecular = vec3(0.0);
		return;*/
		
		rAlbedo *= texture(albedoTexture, texCoord).rgb;
	}
		
	if(useMetallicTexture)
		rMetallic *= texture(metallicTexture, texCoord).r;
		
	if(useRoughnessTexture)
		rRoughness = texture(roughnessTexture, texCoord).r;
		
	if(useNormalTexture){
		//rAlbedo = texture(normalTexture, texCoord).xyz;
		rNormal = texture(normalTexture, texCoord).xyz * 2.0 - 1.0;
		rNormal = normalize(TBN * rNormal); 
	}
	
	// Initialize colors:
    totalDiffuse = ambientLight * rAlbedo;
    totalSpecular = vec3(0.0,0.0,0.0);
	
    // Reflection coefficient if view dir is perpendicular to face:
    vec3 r0 = mix(vec3(0.04), rAlbedo, rMetallic).rgb;
        
    // Add all light sources:
    for(int i=0; i < 4 ; ++i){
		vec4 lightPosition = mLightPositions[i];
	
        // Vector from fragment to light:
        vec3 toLight = normalize(lightPosition.xyz - pPos.xyz*lightPosition.w);
        
        // Calculate light intensity:
        vec3 radiance = vec3(1.0,1.0,1.0) * 3;
        
        float NdotL = max(dot(rNormal,toLight),0.0);
            
        // Attenuation:
        float spotLightFactor = max(dot(mLightDir, -toLight),0.0);
		
        float dist = distance(lightPosition, pPos);
        radiance *= 1.0 / (dist * dist) * spotLightFactor;
                  
        // Half Vector (fragment->light + fragment->camera):
        vec3 halfVec = normalize(toLight + toCamera);
        
        vec3 F = fresnel(max(dot(halfVec, toCamera), 0.0), r0);
        
        vec3 specular = cookTorrance(rNormal, halfVec, toCamera, toLight, F, rRoughness);  
        
        vec3 kD = (vec3(1.0) - F) * (1.0 - pMetallic);
        
		float shadow = ShadowCalculation(i);
		
        
		totalDiffuse += max((kD * rAlbedo / M_PI) * radiance * NdotL, vec3(0.0,0.0,0.0)) * (1 - shadow);
        totalSpecular += specular * radiance * NdotL * (1 - shadow);
    }
	
	
	totalDiffuse = pow(totalDiffuse, vec3(1.0/M_GAMMA));
	totalSpecular = pow(totalSpecular, vec3(1.0/M_GAMMA));
    
    totalSpecular += pEmission;
}

void main()
{
	vec3 totalDiffuse;
	vec3 totalSpecular;
	
	getLighting(mPos, mNormal, color.rgb, emission.rgb, metallic, roughness, totalDiffuse, totalSpecular);	
	
	FragColor = vec4(totalDiffuse + totalSpecular, 1) * 1.6; // 1.5 is just a brightness factor.
}
